{*****************************************************************************}
{*                                                                           *}
{*                              TORPATCH                                     *}
{*                                v1.0                                       *}
{*                                                                           *}
{*         A tool to join a series of points via sections of tori            *}
{*                                                                           *}
{*                     Copyright (C) 1995 M.Mackey.                          *}
{*                                                                           *}
{*****************************************************************************}

program torpatch;
uses crt;
type real=extended;
     vector=record
              x,y,z:real;
            end;
     point=vector;   {Not strictly necessary, but the distinction
                      is sometimes useful}
     pointlistptr=^pointlisttype;
     pointlisttype=record
                     pt:point;
                     next:pointlistptr;
                   end;

var a1,a2,a3:point;
    intangent,outtangent,centre:vector;
    radius,xang,yang:real;
    min,max:vector;
    max_radius:real;
    infile,outfile:text;
    showpoints,showtangents,tangentmode,intersection:boolean;
    pnum:integer;
    s:string;
    name:string[40];
    ch:char;
    num_random,i,repeats:integer;
    pointlisthead:pointlisttype;
    plist:pointlistptr;

function exist(filename:string):boolean;
{Returns true iff file "filename" exists in current directory}
var fil:file;
    a:integer;
begin
  assign(fil,filename);
  {$I-}
  reset(fil);
  {$I+}
  a:=ioresult;
  exist:=(a=0);
  if a=0 then close(fil);
end;

procedure cross(v1,v2:vector;var result:vector);
{Vector cross product}
begin
  result.x:=v1.y*v2.z-v2.y*v1.z;
  result.y:=-(v1.x*v2.z-v2.x*v1.z);
  result.z:=v1.x*v2.y-v2.x*v1.y;
end;

function dot(v1,v2:vector):real;
{Vector dot product}
begin
  dot:=v1.x*v2.x+v1.y*v2.y+v1.z*v2.z;
end;

procedure normalise(var v:vector);
{Normalises a vector}
var r:real;
begin
  r:=sqrt(sqr(v.x)+sqr(v.y)+sqr(v.z));
  if r<>0 then
  begin
    v.x:=v.x/r;
    v.y:=v.y/r;
    v.z:=v.z/r;
  end;
end;

procedure rotatex(v:vector;ang:real;var result:vector);
  {rotate a vector/point around x axis by ang radians}
var sinang,cosang:real;
begin
  sinang:=sin(ang);
  cosang:=cos(ang);
  result.x:=v.x;
  result.y:=v.y*cosang + v.z*sinang;
  result.z:=-v.y*sinang + v.z*cosang;
end;

procedure rotatey(v:vector;ang:real;var result:vector);
  {rotate a vector/point around y axis by ang radians}
var sinang,cosang:real;
begin
  sinang:=sin(ang);
  cosang:=cos(ang);
  result.x:=v.x*cosang + v.z*sinang;
  result.y:=v.y;
  result.z:=-v.x*sinang + v.z*cosang;
end;

procedure addvectors(v1,v2:vector;var result:vector);
{Add two vectors}
begin
  result.x:=v1.x+v2.x;
  result.y:=v1.y+v2.y;
  result.z:=v1.z+v2.z;
end;

procedure subtractvectors(v1,v2:vector;var result:vector);
{Subtract two vectors}
begin
  result.x:=v1.x-v2.x;
  result.y:=v1.y-v2.y;
  result.z:=v1.z-v2.z;
end;

procedure scalevector(var v:vector; scalefactor:real);
{Vector by scalar multiplication}
begin
  v.x:=v.x*scalefactor;
  v.y:=v.y*scalefactor;
  v.z:=v.z*scalefactor;
end;

procedure shiftvector(var v:vector; shift:real);
{scalar addition to vector}
begin
  v.x:=v.x+shift;
  v.y:=v.y+shift;
  v.z:=v.z+shift;
end;

function distance(p1,p2:point):real;
{Returns the distance between two points in 3-space}
begin
  distance:=sqrt( sqr(p1.x-p2.x)+sqr(p1.y-p2.y)+sqr(p1.z-p2.z) );
end;

procedure writevector(heading:string;v:vector);
{Writes a vector to stdout}
begin
  writeln(heading,' x:',v.x:12:6,' y:',v.y:12:6,' z:',v.z:12:6);
end;

procedure Find2DCentre(q1,q2,q3:vector;var result:vector);
{Finds the circumcentre of q1..3 (in two dimensions). Assumes that
 q1..q3 have the same z-coordinate.}
var r1,r2:vector;
    m1,m2:real;
begin
  addvectors(q1,q2,r1);
  scalevector(r1,0.5);
  addvectors(q2,q3,r2);
  scalevector(r2,0.5);   {r1 and r2 are the midpoints of the line segments
                          q1->q2 and q2->q3}

  if (q1.x-q2.x)<>0 then
    m1:=(q1.y-q2.y)/(q1.x-q2.x)   {gradient of first line segment}
  else
    m1:=(q1.y-q2.y)*1.0e10;
  if (q2.x-q3.x)<>0 then
    m2:=(q2.y-q3.y)/(q2.x-q3.x)   {gradient of second line segment}
  else
    m2:=(q2.y-q3.y)*1.0e10;
  if m1=0 then m1:=1.00e-10;
  if m2=0 then m2:=1.01e-10;
  if m1=m2 then m1:=m1*1.000000001;  {should really output cylinder, but..}

  result.x:=( (r2.x/m2 - r1.x/m1) + (r2.y - r1.y) )/(1/m2 - 1/m1);
                                 {basic Cartesian geometry :)}
  result.y:=-1/m1 * (result.x-r1.x) + r1.y;
  result.z:=q1.z;  {q1..3 lie in same z plane, remember?}
end;

procedure Find2DCentre2(q1,q2:point;tangent:vector;var result:vector);
{Finds the circumcentre of q1 and q2 given the tangent vector at q1
 in two dimensions. Assumes that the circumcircle lies in the xy plane
 ie that q1 and q2 have the same z coordinate and that the tangent
 vector has zero z component}
var r1,r2:vector;
    m1,m2:real;
begin
  r1:=q1;
  addvectors(q1,q2,r2);
  scalevector(r2,0.5); {r1 and r2 are the endpoints of two radii}
  if (tangent.x<>0) then
    m1:=(tangent.y/tangent.x)   {gradient of tangent}
  else
    m1:=tangent.y*1.0e10;       {close enough}
  if (q1.x-q2.x)<>0 then
    m2:=(q1.y-q2.y)/(q1.x-q2.x)   {gradient of chord q1->q2}
  else
    m2:=(q1.y-q2.y)*1.0e10;     {close enough}
  if m1=0 then m1:=1.0e-10;     {"}
  if m2=0 then m2:=1.0e-10;     {"}
  if m1=m2 then m1:=m1*1.000000001;  {should really output cylinder, but..}
  result.x:=( (r2.x/m2 - r1.x/m1) + (r2.y - r1.y) )/(1/m2 - 1/m1);
                               {more elementary geometry :) }
  result.y:=-1/m1 * (result.x-r1.x) + r1.y;
  result.z:=q1.z;  {assume that q1..3 lie in same z plane}
end;

procedure FindTorusCentre(p1,p2,p3:point;var centre:point;
   var intangent,outtangent:vector;var radius,xang,yang:real);

{takes 3 points in 3-space as input. Returns the centre and radius of
  a torus in the xy plane and two angles to rotate this torus so that
  it will pass through the original 3 points. Also returns the tangent
  vectors from the first and third points (untransformed ie in original
  coordinate frame)

  The inwards and outgoing tangents are guaranteed to point in opposite
  directions along the torus arc.}

var r1,r2:vector; {r1=p1->p2, r2=p2->p3}
    n:vector;  {normal to plane of p1..3}
    q1,q2,q3:point; {p1..3 rotated so their plane is parallel to xy plane}
    test:vector;
    m:real;
    positive:boolean; {records the relative direction of the inwards
                       and outwards tangents}

begin
  subtractvectors(p2,p1,r1);      {r1 = p1->p2}
  subtractvectors(p3,p2,r2);      {r2 = p2->p3}
  cross(r1,r2,n);                 {normal vector to p1,p2,p3 plane}
  if n.z<>0 then
    xang:=arctan(-n.y/n.z)
  else xang:=arctan(-n.y*1.0e10);
  rotatex(n,xang,n);              {rotate n to eliminate x component}
  if n.z<>0 then
    yang:=arctan(-n.x/n.z)
  else yang:=arctan(-n.x*1.0e10);
  rotatey(n,yang,n);              {rotate n to eliminate y component}
  rotatex(p1,xang,q1);
  rotatey(q1,yang,q1);
  rotatex(p2,xang,q2);
  rotatey(q2,yang,q2);
  rotatex(p3,xang,q3);            {rotate p1..p3 by these angles to}
  rotatey(q3,yang,q3);            {obtain q1..q3 with constant z-coords}
  Find2DCentre(q1,q2,q3,centre);  {Find circumcentre in 2D}
  radius:=distance(q1,centre);    {Radius of torus}

  m:=(centre.y-q1.y)/(centre.x-q1.x); {gradient of 2D vector from centre to
                                       1st point}
  intangent.x:=1;
  if m<>0 then
    intangent.y:=-1/m
  else
    intangent.y:=-1.0e10;
  intangent.z:=0;
  rotatey(intangent,-yang,intangent);
  rotatex(intangent,-xang,intangent); {transform tangent vector to original
                                       coordinate frame}

  subtractvectors(q1,centre,test);    {test = centre -> q1}
  cross(test,intangent,test);         {which direction is this tangent
                                       pointing?}
  positive:=test.z>0;   {for comparison with outtangent: are they
                         pointing in same/opp dir?}

  m:=(centre.y-q3.y)/(centre.x-q3.x); {gradient of 2D vector from centre to
                                       3rd point}
  outtangent.x:=1;
  outtangent.y:=-1/m;

  outtangent.z:=0;
  subtractvectors(q3,centre,test);    {test = centre -> q3}

  cross(test,outtangent,test);        {which direction is this tangent
                                       pointing?}
  if (positive and (test.z>0)) or (not positive and (test.z<0)) then
    scalevector(outtangent,-1);
      {make sure intangent and outtangent point in opp. dirs}

  rotatey(outtangent,-yang,outtangent);
  rotatex(outtangent,-xang,outtangent); {transform tangent vector to original
                                         coordinate frame}
end;

procedure FindTorusCentre2(p1,p2:point;intangent:vector;var centre:point;
                       var outtangent:vector;var radius,xang,yang:real);
{takes 2 points in 3-space and an initial tangent vector (pointing away
  from the arc) as input. Returns the centre and radius of a torus in
  the xy plane and two angles to rotate this torus so that it will pass
  through the original 3 points. Also returns the tangent vector from
  the second point (untransformed ie in original coordinate frame).

  The inwards and outgoing tangents are guaranteed to point in opposite
  directions along the torus arc.}

var r:vector; {r1=p1->p2}
    n:vector;  {normal to plane of p1->p2 and intangent vector}
    q1,q2:point; {transformed p1,p2 into xy plane}
    test:vector;
    m:real;
    positive:boolean; {records the relative direction of the inwards
                       and outwards tangents}


begin
  subtractvectors(p2,p1,r);   {r = p1->p2}
  cross(intangent,r,n);       {normal vector to plane of intangent and r}
  if n.z<>0 then
    xang:=arctan(-n.y/n.z)
  else
    xang:=arctan(-n.y*1.0e10);
  rotatex(n,xang,n);          {rotate normal to eliminate x component}
  if n.z<>0 then
    yang:=arctan(-n.x/n.z)
  else
    yang:=arctan(-n.x*1.0e10);
  rotatey(n,yang,n);          {rotate normal to eliminate y component}
  rotatex(p1,xang,q1);
  rotatey(q1,yang,q1);
  rotatex(p2,xang,q2);        {rotate p1 and p2 by these angles to}
  rotatey(q2,yang,q2);        {obtain q1 and q2 with constant z-coords}
  rotatex(intangent,xang,intangent);
  rotatey(intangent,yang,intangent);  {rotate intangent into same coordinate
                                       frame}

  Find2DCentre2(q1,q2,intangent,centre); {find circumcentre in 2 dimensions}
  radius:=distance(q1,centre);         {radius of circle}

  subtractvectors(q1,centre,test);  {test = centre -> q1}
  cross(test,intangent,test);       {what direction is this tangent pointing?}
  positive:=test.z>0;   {for comparison with outtangent: pointing in
                         same/opp dir?}

  m:=(centre.y-q2.y)/(centre.x-q2.x); {gradient of radius to 2rd point}
  outtangent.x:=1;
  outtangent.y:=-1/m;
  outtangent.z:=0;
  subtractvectors(q2,centre,test);  {vector from centre to endpoint}
  cross(test,outtangent,test);      {what direction is this tangent pointing?}

  if (positive and (test.z>0)) or (not positive and (test.z<0)) then
    scalevector(outtangent,-1);
      {make sure intangent and outtangent point in opp. dirns}

  rotatey(outtangent,-yang,outtangent);
  rotatex(outtangent,-xang,outtangent); {transform to original coord frame}

end;

procedure Paramhelp;
{Print help on TORPATCH syntax}
begin
  writeln;
  writeln('Syntax is TORPATCH [-t] [-p] [-c] [-n name] <input file> <output file>');
  writeln;
  writeln('The -t flag signifies that the first two points define a');
  writeln('  tangent to the first torus. If this flag is not specified');
  writeln('  then the first torus patch is defined to pass through the');
  writeln('  first three points.');
  writeln;
  writeln('The -p flag indicates that a structure called Points containing');
  writeln('  the control points is to be included in the output.');
  writeln;
  writeln('The -c flag indicates that tangent cylinders are to be');
  writeln('  placed at each control point. The red cylinders point');
  writeln('  in the ''forwards'' direction. This option is mainly');
  writeln('  included for debugging purposes.');
  writeln;
  writeln('The -n flag allows you to specify the name of the torus patch. The');
  writeln('  default name is Torus_Patch. Textures have ''_Texture'' appended.');
  halt(0);
end;

procedure OutputTangent(a1:point;intangent:vector;a3:point;outtangent:vector);
{Print tangent object}
begin
  normalise(outtangent);
  normalise(intangent);
  writeln(outfile,'    object{');
  writeln(outfile,'      cylinder{');
  writeln(outfile,'        <',a1.x:12:6,',',a1.y:12:6,',',a1.z:12:6,'>');
  writeln(outfile,'        <',a1.x:12:6,',',a1.y:12:6,',',a1.z:12:6,'> +');
  writeln(outfile,'          (<',intangent.x:12:6,',',intangent.y:12:6,',',outtangent.z:12:6,'> * Tangent_Length)');
  writeln(outfile,'      Tangent_Radius');
  writeln(outfile,'      texture{Tangent_Texture_In}');
  writeln(outfile,'      }');
  writeln(outfile,'    }');
  writeln(outfile,'    object{');
  writeln(outfile,'      cylinder{');
  writeln(outfile,'        <',a3.x:12:6,',',a3.y:12:6,',',a3.z:12:6,'>');
  writeln(outfile,'        <',a3.x:12:6,',',a3.y:12:6,',',a3.z:12:6,'> +');
  writeln(outfile,'          (<',outtangent.x:12:6,',',outtangent.y:12:6,',',outtangent.z:12:6,'> * Tangent_Length)');
  writeln(outfile,'        Tangent_Radius');
  writeln(outfile,'        texture{Tangent_Texture_Out}');
  writeln(outfile,'      }');
  writeln(outfile,'    }');
end;

function snip(var s:string):string;
begin
  while (length(s)>0) and ((s[1]=' ') or (s[1]=',')) do s:=copy(s,2,length(s));
  i:=0;
  repeat
    inc(i)
  until (i>length(s)) or (s[i]=' ') or (s[i]=',');
  snip:=copy(s,1,i-1);
  s:=copy(s,i,length(s));
end;

function readcoords(var f:text;var x,y,z:real;var s:string):boolean;
  {returns success value}
var s2,s3:string;
    i:integer;
begin
  readcoords:=true;
  readln(f,s2);
  s:=s2;
  s3:=snip(s);
  val(s3,x,i);
  if i<>0 then readcoords:=false;
  s3:=snip(s);
  val(s3,y,i);
  if i<>0 then readcoords:=false;
  s3:=snip(s);
  val(s3,z,i);
  if (i<>0) or (s3='') then readcoords:=false;
  s:=s2;
end;

function GetRandomPoint(var a2:vector;min,max,a1:vector;limit:real):boolean;
{returns a point within limit of a1, inside box <min, max> if possible}
var temp:vector;
begin
  a2.x:=random*(max.x-min.x)+min.x;
  a2.y:=random*(max.y-min.y)+min.y;
  a2.z:=random*(max.z-min.z)+min.z;
  GetRandomPoint:=false;
  temp:=min;
  shiftvector(temp,-limit);
  if (temp.x<a1.x) and (temp.y<a1.y) and (temp.z<a1.z) then
  begin
    temp:=max;
    shiftvector(temp,limit);
    if (temp.x>a1.x) and (temp.y>a1.y) and (temp.z>a1.z) then
    begin
      i:=0;
      repeat
        inc(i);
        a2.x:=random*(max.x-min.x)+min.x;
        a2.y:=random*(max.y-min.y)+min.y;
        a2.z:=random*(max.z-min.z)+min.z;
      until (i>10000) or (distance(a2,a1)<=limit);
      if i<10000 then GetRandomPoint:=true;
    end;
  end;
end;

procedure addtopointlist(p:point);
begin
  plist^.pt:=p;
  new(plist^.next);
  plist^.next^.next:=nil;
  plist:=plist^.next;
end;


procedure OutputTorus(a1,a2,centre:point;radius,xang,yang:real;intangent:vector;
                      intersection:boolean);
{Prints one torus patch object from two endpoints, centre, rotation angles
 and ingoing tangent. The intersection parameter controls the clipping planes
 and depends on whether the arc of the torus used subtends more than
 180 degrees}
begin
  writeln(outfile,'    object{');
  writeln(outfile,'      torus{ ',radius:12:6,', Minor_Radius}');
  writeln(outfile,'      rotate x*90');
  writeln(outfile,'      bounded_by{');
  writeln(outfile,'        box{');
  writeln(outfile,'          <',-radius*1.001:6:3,'-Minor_Radius, ',-radius*1.001:6:3,
                              '-Minor_Radius, -Minor_Radius*1.001>');
  writeln(outfile,'          <',radius*1.001:6:3,'+Minor_Radius, ',radius*1.001:6:3,
                              '+Minor_Radius, Minor_Radius*1.001>');
  writeln(outfile,'        }');
  writeln(outfile,'      }');
  writeln(outfile,'      translate< ',centre.x:12:6,',',centre.y:12:6,',',centre.z:12:6,'>');
  writeln(outfile,'      rotate y*(',-yang*180/pi:12:6,')');
  writeln(outfile,'      rotate x*(',xang*180/pi:12:6,')');

  {Note: the change in sign in xang is because I am using a right-handed
     coordinate system within this program, but the output is for the
     POV default left-handed coordinate system. The rotation in the
     x direction thus has to be reversed.}

  writeln(outfile,'      texture{',name,'_Texture}');
  writeln(outfile,'      clipped_by{');
  if intersection then
    writeln(outfile,'        intersection{')
  else
    writeln(outfile,'        union{');
  writeln(outfile,'          plane{ <',intangent.x:12:6,',',intangent.y:12:6,',',intangent.z:12:6,'> 0');
  writeln(outfile,'            translate <',a1.x:12:6,',',a1.y:12:6,',',a1.z:12:6,'>}');
  writeln(outfile,'          plane{ <',outtangent.x:12:6,',',outtangent.y:12:6,',',outtangent.z:12:6,'> 0');
  writeln(outfile,'            translate <',a2.x:12:6,',',a2.y:12:6,',',a2.z:12:6,'>}');
  writeln(outfile,'        }');
  writeln(outfile,'      }');
  writeln(outfile,'    }');
end;

begin
  randseed:=2;
  if paramcount<2 then ParamHelp;
  tangentmode:=false;
  showtangents:=false;
  showpoints:=false;
  name:='Torus_Patch';
  pnum:=1;
  plist:=@pointlisthead;
  num_random:=10000;
  s:=paramstr(pnum);
  while (s[1]='-') or (s[1]='/') do   {parse parameters}
  begin
    case upcase(s[2]) of
      'T':tangentmode:=true;
      'C':showtangents:=true;
      'P':showpoints:=true;
      'N':begin
            name:=paramstr(pnum+1);
            inc(pnum);
          end;
      else ParamHelp;
    end;
    inc(pnum);
    s:=paramstr(pnum);
  end;
  if paramcount<>pnum+1 then ParamHelp;
  assign(infile,paramstr(pnum));
{$I-}
  reset(infile);
  if ioresult<>0 then
  begin
    writeln('File ',paramstr(pnum),' not found.');
    ParamHelp;
  end;
  assign(outfile,paramstr(pnum+1));
  if exist(paramstr(pnum+1)) then
  begin
    writeln('File ',paramstr(pnum+1),' exists. Overwrite (y/n)?');
    repeat
      ch:=upcase(readkey);
    until (ch='Y') or (ch='N');
    if ch='N' then halt(0);
  end;
  rewrite(outfile);
  writeln('Joining points:');
  repeat until readcoords(infile,a1.x,a1.y,a1.z,s) or eof(infile);
  repeat until readcoords(infile,a2.x,a2.y,a2.z,s) or eof(infile);
  repeat until readcoords(infile,a3.x,a3.y,a3.z,s) or eof(infile);
  if TangentMode then
  begin
    addtopointlist(a2);
    addtopointlist(a3);
    writevector('Initial tangent control point: ',a1);
    writevector('Initial tangent at: ',a2);
  end
  else
  begin
    addtopointlist(a1);
    addtopointlist(a2);
    addtopointlist(a3);
    writevector('Initial control point',a1);
    writevector('Initial control point',a2);
  end;
  writevector('Initial control point',a3);
  if not TangentMode then
  begin
    FindTorusCentre(a1,a2,a3,centre,intangent,outtangent,radius,xang,yang);
  end
  else
  begin
    subtractvectors(a1,a2,intangent);
    FindTorusCentre2(a2,a3,intangent,centre,outtangent,radius,xang,yang);
  end;
  writeln(outfile,'/*-----------------------------------------------------------------------\');
  writeln(outfile,'|  This file was generated by TORPATCH v1.0 copyright 1995 Mark Mackey   |');
  writeln(outfile,'\-----------------------------------------------------------------------*/');
  writeln(outfile);
  writeln(outfile,'#declare Minor_Radius=0.05');
  writeln(outfile,'#declare ',name,'_Texture=texture{pigment{Green}}');
  writeln(outfile);
  if ShowPoints then
  begin
    writeln(outfile,'#declare Point_Radius=0.1');
    writeln(outfile,'#declare Point_Texture=texture{pigment{White}}');
    writeln(outfile);
  end;
  if ShowTangents then
  begin
    writeln(outfile,'#declare Tangent_Radius=0.05');
    writeln(outfile,'#declare Tangent_Length=1');
    writeln(outfile,'#declare Tangent_Texture_In=');
    writeln(outfile,'  texture{');
    writeln(outfile,'    pigment{Red}');
    writeln(outfile,'  }');
    writeln(outfile,'#declare Tangent_Texture_Out=');
    writeln(outfile,'  texture{');
    writeln(outfile,'    pigment{Blue}');
    writeln(outfile,'  }');
    writeln(outfile);
  end;
  writeln(outfile,'#declare ',name,' =');
  writeln(outfile,'  union{');
  intersection:=((intangent.x*(a2.x-a1.x)+intangent.y*(a2.y-a1.y)+
    intangent.z*(a2.z-a1.z))<0);
  if tangentmode then
    OutputTorus(a2,a3,centre,radius,xang,yang,intangent,intersection)
  else
    OutputTorus(a1,a3,centre,radius,xang,yang,intangent,intersection);
  if showtangents then
  begin
    OutputTangent(a1,intangent,a3,outtangent);
  end;
  a1:=a3;
  intangent:=outtangent;
  scalevector(intangent,-1);
  if eoln(infile) and not eof(infile) then readln(infile);
  while (not eof(infile)) do
  begin
    if readcoords(infile,a2.x,a2.y,a2.z,s) then
    begin
      addtopointlist(a2);
      writevector('Control point: ',a2);
      if eoln(infile) and not eof(infile) then readln(infile);
      FindTorusCentre2(a1,a2,intangent,centre,outtangent,radius,xang,yang);
      intersection:=((intangent.x*(a2.x-a1.x)+intangent.y*(a2.y-a1.y)+
        intangent.z*(a2.z-a1.z))<0);
      OutputTorus(a1,a2,centre,radius,xang,yang,intangent,intersection);
      if showtangents then
      begin
        OutputTangent(a1,intangent,a2,outtangent);
      end;
      a1:=a2;
      intangent:=outtangent;
      scalevector(intangent,-1);
    end
    else
    if (length(s)>0) and (upcase(s[1])='R') then
    begin
      readln(infile,num_random,randseed,max_radius);
      readln(infile,min.x,min.y,min.z);
      readln(infile,max.x,max.y,max.z);
      repeat
        dec(num_random);
        repeats:=0;
        repeat
          inc(repeats);
          if (repeats mod 10)=0 then write('.');
          GetRandomPoint(a2,min,max,a1,2*max_radius);
          FindTorusCentre2(a1,a2,intangent,centre,outtangent,radius,xang,yang);
        until (radius<max_radius) or (repeats>500);
        if repeats>500 then
        begin
          writeln;
          writeln('WARNING: Unable to meet maximum radius specification.');
          writeln('         Radius used: ',radius:12:6);
        end;
        addtopointlist(a2);
        writevector('Random control point: ',a2);
        intersection:=((intangent.x*(a2.x-a1.x)+intangent.y*(a2.y-a1.y)+
          intangent.z*(a2.z-a1.z))<0);
        OutputTorus(a1,a2,centre,radius,xang,yang,intangent,intersection);
        if showtangents then
        begin
          OutputTangent(a1,intangent,a2,outtangent);
        end;
        a1:=a2;
        intangent:=outtangent;
        scalevector(intangent,-1);
      until (num_random=0);
    end
    else
    begin
      while (length(s)>0) and ((s[1]=' ') or (s[1]=',')) do s:=copy(s,2,length(s));
      if s<>'' then
      begin
        writeln('ERROR: unable to parse input line.');
        close(infile);
        close(outfile);
        halt;
      end;
    end;
  end;
  writeln(outfile,'  }');
  writeln(outfile);
  if ShowPoints then
  begin
    writeln(outfile,'#declare Points =');
    writeln(outfile,'  union{');
    plist:=@pointlisthead;
    repeat
      with plist^ do
      writeln(outfile,'   sphere{ <',pt.x:12:6,', ',pt.y:12:6,', ',pt.z:12:6,'> Point_Radius');
      writeln(outfile,'     texture{Point_Texture}');
      writeln(outfile,'   }');
      plist:=plist^.next;
    until plist^.next=nil;
    writeln(outfile,'  }');
  end;
  close(outfile);
  close(infile);
  writeln;
  writeln('Patch file written successfully.');
  writeln('Thanks for using TORPATCH by Mark Mackey!');
end.
